<h4>SaltStack News</h4>
<p><a href="http://www.saltstack.com/events">Upcoming SaltStack events,
webinars and local meet ups and user groups.</a></p>

{% if on_saltstack %}
<h4>SaltStack training</h4>
<a href="http://www.saltstack.com/training/">
<img src="{{ pathto('_static/images/salt-training.png', 1) }}" alt="" style="padding-bottom: 1em;">
</a>
{% endif %}
