# This file was auto-generated by salt's setup on Friday, 03 October 2014 @ 20:10:16 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2014, 7, 0, 0, 3, 0, None)
